/*
** uri.c - part of GrabURL 2.0
** Copyright (C) 1996-97 Serge Emond
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "uri.h"
#include "log.h"
#include "config.h"
#include "macros.h"
#include "libgurl.h"

struct URI_Head uri_list;

/*
** Adds an URI to the list.  If flags == -2, master's flags are used.
** If depth == -2, master's depth is used.
** If master == 0, then defaults are used.
*/

int AddURI(struct URI *master, char *uri, int flags, int depth,
		   const char *saveto)
{
	struct URI *new;
	struct SURI *sur, *sur2;
	char *stt;

	sur = SplitURI(uri);
	/*	if (!sur) return(0); */
	if (FindURI((sur ? sur->nuri : uri), -1, -1)) return(0);
	if (master) {
	  sur2 = SplitURI(master->uri);
	  if (!sur2) return(0);

	  if ((flags !=-2) && (flags & URI_STAYINHOST) &&
		  (str_ieq(sur->host, sur2->host) != 0)) return(0);
	  FreeSURI(sur2);
	}

	new = (struct URI *)gmalloc(sizeof(struct URI));

	new->uri = (char *)gmalloc(strlen((sur ? sur->nuri : uri))+1);
	strcpy(new->uri, (sur ? sur->nuri : uri));

	if (!sur)
	  new->flags = URI_SKIP;
	else {
	  FreeSURI(sur);
	  new->flags = (flags==-2 ? (master ? master->flags : cfg_global.def_flags)
					: flags);
	}
	new->depth = (depth==-2 ? (master ? master->depth
							   : cfg_global.def_depth) : depth);

	if (saveto) stt = (char *)saveto;
	else if (cfg_global.def_saveto) stt = cfg_global.def_saveto;
	else stt = 0;

	if (!stt) new->saveto = 0;
	else {
	  new->saveto = (char *)gmalloc(strlen(stt)+1);
	  strcpy(new->saveto, stt);
	}

	new->prev = uri_list.last;
	if (uri_list.last) uri_list.last->next = new;
	uri_list.last = new;
	new->next = 0;
	if (!uri_list.first) uri_list.first = new;

	return(1);
}

struct URI *FindURI(char *uri, int reqflags, int excflags)
{
	struct URI *p;

	for (p = uri_list.first; p; p = p->next) {
	  if (uri && str_ieq(uri, p->uri)) continue;
	  if ((reqflags != -1) && ((p->flags & reqflags) != reqflags)) continue;
	  if ((excflags != -1) && ((p->flags & excflags) != 0)) continue;
	  return(p);
	}
	return(0);
}

int CountURI(int reqflags, int excflags)
{
	struct URI *p;
	int cnt;

	cnt = 0;

	for (p = uri_list.first; p; p = p->next) {
	  if (reqflags >= 0 && ((p->flags & reqflags) != reqflags)) continue;
	  if (excflags >= 0 && ((p->flags & excflags) != 0)) continue;
	  cnt++;
	}

	return cnt;
}

int SaveURIFile(const char *fname)
{
    FILE *fh;
	struct URI *p;
	char *line;

	fh = fopen(fname, "w");
	if (!fh) {
	  fprintf(stderr, "SaveURIFile: can't open file\n");
	  exit(EXIT_FAILURE);
	}

	for (p = uri_list.first; p; p = p->next) {
	  line = MakeURILine(p);
	  fprintf(fh, "%s\n", line);
	}

	fclose(fh);

	return(0);
}

int LoadURIFile(const char *fname)
{
	FILE *fh;
	char line[MAX_UFILE_LLEN], *p;
	int ret, got;

	fh = fopen(fname, "r");
	if (!fh) {
	  fprintf(stderr, "LoadURIFile: can't open file\n");
	  exit(EXIT_FAILURE);
	}

	got = 0;

	while (fsgetdelim(fh, line, MAX_UFILE_LLEN, 0, "\n") > 0) {
	  if ((p = strchr(line, '\n')) != 0) {
		p[0] = 0;
	  }
	  p = NextNonBlank(line);
	  if ((p[0] != 0) && p[0] != '#') { /* if line non-blank & non-comment */
		ret = AddURILine(p);
		if (ret < 0) {
		  putlog("Error reading URIFile", LOG_ERROR);
		  return(0);
		}
		else if (ret > 0) {
		  got++;
		}
	  }
	}

	fclose(fh);

	return(got);
}

char *MakeURILine(struct URI *u)
{
    char *line, type;
	char tmp[128];

	line = (char *)gmalloc(MAX_UFILE_LLEN);

	if (u->flags & URI_RECEIVED) type='R';
	else if (u->flags & URI_SKIP) type = 'S';
	else if (u->flags & URI_FAILED) type = 'F';
	else if (u->flags & URI_ERROR) type = 'X';
	else if (u->flags & URI_MOVED) type = 'M';
	else type = 'Q';

	sprintf(line, "%c %s", type, u->uri);

	if (u->saveto) {
	  sprintf(tmp, " TO=\"%s\"", u->saveto);
	  strcat(line, tmp);
	}

	if (u->depth != -1) {
	  sprintf(tmp, " D=%u", u->depth);
	  strcat(line, tmp);
	}

	if (u->flags & URI_RETRY) {
	  strcat(line, " RETRY");
	}
	if (u->flags & URI_NODIRS) {
	  strcat(line, " NODIRS");
	}
	if (u->flags & URI_SCAN) {
	  strcat(line, " SCAN");
	}
	if (u->flags & URI_STAYINHOST) {
	  strcat(line, " STAYONHOST");
	}
	/* Not implemented
	if (u->flags & URI_ADDHEAD) {
	  strcat(line, " ADDHEAD");
	}
	*/
	if (u->flags & URI_SAVEHEAD) {
	  strcat(line, " SAVEHEAD");
	}
	if (u->flags & URI_NONEXISTS) {
	  strcat(line, " NONEXISTS");
	}
	if (u->flags & URI_IFMODIFIED) {
	  strcat(line, " IFMODIFIED");
	}

	return(line);
}

int AddURILine(const char *line)
{
	char *p, *uri_str, *saveto;
	int flags, depth, i, ret;

	if (!isspace(line[1])) {
	  /* We assume this line contains a single URI */
	  return(AddURI(0, (char *)line, -2, -2, 0));
	}

	/* Else, it's a complete line, with type, etc.. */

	switch (line[0]) {
	case 'q':
	case 'Q':
	  flags = 0;
	  break;
	case 'r':
	case 'R':
	  flags = URI_RECEIVED;
	  break;
	case 'f':
	case 'F':
	  flags = URI_FAILED;
	  break;
	case 'x':
	case 'X':
	  flags = URI_ERROR;
	  break;
	case 'm':
	case 'M':
	  flags = URI_MOVED;
	  break;
	default:
	  flags = URI_SKIP;
	}

	uri_str = ExtractArgument(line, 1);
	if (!uri_str) return(-1);

	i = 1; depth = -1; saveto=0;

	while ((p = ExtractArgument(line, ++i)) != 0) {
	  if (!str_nieq(p, "TO=", 3)) saveto = ExtractArgument(p+3, 0);
	  else if (!str_nieq(p, "D=", 2)) depth = strtol(p+2, 0, 10);
	  else if (!str_ieq(p, "RETRY")) flags |= URI_RETRY;
	  else if (!str_ieq(p, "NODIRS") || !str_ieq(p, "ND")) flags |= URI_NODIRS;
	  else if (!str_ieq(p, "NONEXISTS") || !str_ieq(p, "NE"))
		flags |= URI_NONEXISTS;
	  else if (!str_ieq(p, "SCAN")) flags |= URI_SCAN;
	  else if (!str_ieq(p, "STAYONHOST")) flags |= URI_STAYINHOST;
	  /*	  else if (!str_ieq(p, "ADDHEAD")) flags |= URI_ADDHEAD; */
	  else if (!str_ieq(p, "IFMODIFIED")) flags |= URI_IFMODIFIED;

	  free(p);
	}

	ret = AddURI(0, uri_str, flags, depth, saveto);
	free(uri_str);
	return(ret);
}

/* on error, this doesn't return... */

struct SURI *SplitURI(const char *uri)
{
	char *p, *q, *r, *port=0;
	struct SURI *suri;
	int i, len;

	suri = (struct SURI *)gmalloc(sizeof(struct SURI));

	p = strstr(uri, ":");
	if (!p) {
	  /* no protocol - assume http */
	  suri->protocol = (char *)gmalloc(5);
	  strcpy(suri->protocol, "http");
	  len = 4;
	  p = (char *)uri;
	}
	else {
	  if (str_nieq(uri, "http", 4)) {
		/* Unsupported protocol */
		return(0);
	  }
	  i = (int) p - (int) uri;
	  suri->protocol = (char *)gmalloc(i + 1);
	  
	  strncpy(suri->protocol, uri, i);
	  suri->protocol[i] = 0;
	  len = i;
	  p+=3;
	}

	q = strchr(p, '/');
	if (!q) q = strchr(p, '\0');

	r = strchr(q+1, ':');
	if (!r) {
	  /* No port number */
	  r = q;
	}

	i = (int)r - (int)p;
	suri->host = (char *)gmalloc(i+1);
	strncpy(suri->host, p, i);
	suri->host[i] = 0;
	len += i;

	if (r == q) {
	  suri->port = 0;
	}
	else {
	  suri->port = strtol(r+1, 0, 10);
	  i = (int)q - (int)r;
	  len += i;
	  port = (char *)gmalloc(i+1);
	  strncpy(port, r, i);
	  port[i]=0;
	}

	if (q[0]==0) {
	  /* no path */
	  suri->path = 0;
	}
	else {
	  q++;
	  p = strrchr(q, '/');
	  if (!p) {
		/* no path */
		suri->path = 0;
	  }
	  else {
		i = (int) p - (int) q + 1;
		suri->path = (char *)gmalloc(i+1);
		strncpy(suri->path, q, i);
		suri->path[i] = 0;
		q = p+1;
		len += i;
	  }
	}

	/* At this point, q points to the beginning of the filename */

	if (q[0]==0) {
	  suri->filename = 0;
	}
	else {
	  i = strlen(q);
	  suri->filename = (char *)gmalloc(i+1);
	  strcpy(suri->filename, q);
	  len += i;
	}

	/* Ok - all full - build nuri */

	len += 5;					/* "://", "/" (after host) & '\0'*/

	suri->nuri = (char *)gmalloc(len);

	strcpy(suri->nuri, suri->protocol);
	strcat(suri->nuri, "://");
	strcat(suri->nuri, suri->host);
	if (suri->port) {
	  strcat(suri->nuri, port);
	  free(port);
	}
	strcat(suri->nuri, "/");
	if (suri->path) strcat(suri->nuri, suri->path);
	if (suri->filename) strcat(suri->nuri, suri->filename);

	return(suri);
}
